# -*- coding:utf-8 -*-
# Created by qinwei on 2017/12/7
#
import threading
import uuid
from time import time


class FileUtils:
    @staticmethod
    def create_tmp_file_name(prefix="", suffix=""):
        """
        create a new temp file name
        :param prefix:
        :param suffix:
        :return:
        """
        name = "{}{}-{}".format(int(time()), threading.current_thread().ident, uuid.uuid4().hex[:8])
        file_name = "{prefix}{name}{suffix}".format(prefix=prefix, name=name, suffix=suffix)
        return file_name



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/11/24
#

from hunters.constant import Const
from hunters.cookie import MemoryCookieStore


class UserAgent(Const):
    """常用UA"""
    CHROME = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36"
    IPHONE = "Mozilla/5.0 (iPhone; CPU iPhone OS 9_1 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13B143 Safari/601.1"
    ANDROID = "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Mobile Safari/537.36"


class BrowserConfig(object):
    """ 统一浏览器配置项, 统一抽象的描述实现浏览器所具备的一些特性"""
    LOCAL_STORE = "/data/hunter/"

    def __init__(self,
                 user_agent=None,
                 local_store=None,
                 image=True,
                 headless=True,
                 execute_bin=None,
                 cookie_store=MemoryCookieStore(),
                 max_body_size=-1,
                 width=1440,
                 height=1000,
                 gpu=False,  # if in gui, can use gpu
                 extra_options=[],
                 manual_url="",  # manual to connect to current open chrome, not auto start..
                 # manual_url="127.0.0.1:9222"
                 disk_cache_size=None):
        """ The Common base Browser Config """

        self.user_agent = user_agent or UserAgent.CHROME

        self.local_store = local_store or BrowserConfig.LOCAL_STORE

        self.image = image

        self.headless = headless

        #: 分布式Cookie/会话保持, 在分布式环境中, 每产生的URL都插入中央队列, 每个爬虫(可能多台机)从中央队列里取出来的内容
        #: 是无状态的, 如果需要维持爬虫的上下文关系, 需要保证有分布式的session维持cookie内容,
        #: 这个cookie可以集中存储.
        self.cookie_store = cookie_store

        self.max_redirects = 3

        self.max_body_size = max_body_size

        self.disk_cache_size = disk_cache_size  # in bytes

        self.execute_bin = execute_bin

        self.height = height

        self.width = width

        self.gpu = gpu

        self.extra_options = extra_options

        self.manual_url = manual_url


DEFAULT_BROWSER_CONFIG = BrowserConfig()



# -*- coding:utf-8 -*-
# Created by qinwei on 2018/2/26
#

# 根据interval秒数重复执行的定时器
# @setInterval(3, daemon=False)
# def show():
#    print("heelo")
#
# stop = show() # running ...
# print("ll")
# stop.set() # stopping ...
#
import threading


def setInterval(interval, daemon=False):
    def decorator(function):
        def wrapper(*args, **kwargs):
            stopped = threading.Event()

            def loop():  # executed in another thread
                while not stopped.wait(interval):  # until stopped
                    function(*args, **kwargs)

            t = threading.Thread(target=loop)
            t.daemon = daemon  # stop if the program exits
            t.start()
            return stopped

        return wrapper

    return decorator



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/11/13
#
import atexit
import logging
import queue
import threading
from time import sleep, time

from hunters.config import DEFAULT_BROWSER_CONFIG

logger = logging.getLogger("hunters.pool")


class BrowserPool(object):
    """
    浏览器池
    用一个池来装连接, 提供给Spider
    """

    def __init__(self, browser_config=DEFAULT_BROWSER_CONFIG, init_count=5, max_count=20, min_idle=5, timeout=3):
        self._browser_config = browser_config
        self._init_count = init_count
        self._max_count = max_count
        self._min_idle = min_idle
        self._timeout = timeout
        self._pool = list()
        self._current_used_count = 0
        self._lock = threading.Lock()

        self.init_pool()

    def init_pool(self):
        for item in range(self._init_count):
            self._pool.append(self._browser_config.new_browser())

    def get(self):
        browser_ = None
        with self._lock:
            if len(self._pool) > 0:
                browser_ = self._pool.pop(0)
            elif self._current_used_count < self._max_count:
                browser_ = self._browser_config.new_browser()
            self._current_used_count += 1
        return browser_

    def return_resource(self, browser):
        self._pool.append(browser)
        self._current_used_count -= 1

    def close_all(self):
        for item in self._pool:
            item.close()
            item.quit()


class ResourcePool(object):
    """
    一个抽象的池, 对实现Factory接口的池
    """

    def __init__(self, factory=None, init_count=5, max_count=20, min_idle=5, timeout=3, init_pool=False):
        self._factory = factory
        self._init_count = init_count
        self._max_count = max_count
        self._min_idle = min_idle
        self._timeout = timeout
        self._pool = queue.Queue()
        self._current_used_count = 0
        self._resource_config = factory.config()
        self._lock = threading.Lock()
        self._init_lock = threading.RLock()
        self._is_init = init_pool
        self._resource_use = dict()
        self._resource_map = dict()

    def init_pool(self):
        for item in range(self._init_count):
            self._pool.put(self.new_instance())

        atexit.register(self.close_all)

    def new_instance(self):
        return self._factory.new_instance()

    def get(self):
        with self._init_lock:
            if not self._is_init:
                self.init_pool()
                self._is_init = True

        while True:
            resource = self._get()
            if resource is None:
                print("Wait For Resource thread[{}]".format(threading.current_thread().ident))
                sleep(1)
            else:
                return resource

    def _get(self):
        resource = None
        with self._lock:
            try:
                resource = self._pool.get_nowait()
            except queue.Empty:
                if self._current_used_count < self._max_count:
                    resource = self.new_instance()
                else:
                    return resource

            self._current_used_count += 1

            resource_id = self.___id_key(resource)
            self._resource_use.setdefault(resource_id, {"count": 0, "resource": resource, "last_time": time()})
            self._resource_use[resource_id]["count"] += 1
            self._resource_use[resource_id]["last_time"] = time()
            # for k, v in self._resource_use.items():
            #     logger.debug("ID:{}, COUNT:{}".format(k, v.get("count")))
            return resource

    @staticmethod
    def ___id_key(resource):
        return "id " + str(id(resource))

    def return_resource(self, browser):
        with self._lock:
            self._pool.put(browser)
            self._current_used_count -= 1

    def return_broken_resource(self, browser):
        # close broken resource
        self._resource_use.pop(self.___id_key(browser))
        self.close_item(browser)
        # create a new one to pool
        self.return_resource(self.new_instance())

    def close_all(self):
        self._factory.destroy_all(self._pool)

    def close_item(self, instance):
        self._factory.destroy(instance)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/11/16
import html
import re

from lxml import etree


class HtmlUtils:
    RE_URL = re.compile(r"""url=(.*)""", re.I)

    @staticmethod
    def get_meta_refresh_url(element):
        """  获取refresh隐藏的刷新URL地址
         <meta http-equiv="refresh" content="0;url=http://www.baidu.com/">
         """
        if HtmlUtils.equal_tag_name(element, "meta"):
            content = element.get("content", "")
            match = HtmlUtils.RE_URL.search(content)
            if match:
                return match.group(1).strip()
        return None

    @staticmethod
    def text(element):
        result = []
        if isinstance(element, (list, tuple)):
            for el in element:
                result.append("".join([x for x in el.itertext()]))
            return "".join(result)

        return "".join([x for x in element.itertext()])

    @staticmethod
    def html(element):
        return etree.tostring(element).decode()

    @staticmethod
    def decode_html(html_str):
        return html.unescape(html_str)

    @staticmethod
    def equal_tag_name(element, tag_name):
        return hasattr(element, "tag") and element.tag == tag_name



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/8
import unittest
from concurrent.futures import ThreadPoolExecutor
from time import sleep, time

from hunters.browser import ChromeTab, BaseChrome
from hunters.chrome import ChromeTabDevToolFactory
from hunters.config import BrowserConfig
from hunters.factory.browser_factory import ViewBrowserFactory
from hunters.pool import ResourcePool


class TestBrowser(unittest.TestCase):
    def setUp(self):
        canary = r"C:\Users\ADMIN\AppData\Local\Google\Chrome SxS\Application\chrome.exe"
        # chrome = r"C:\Program Files (x86)\Google\Chrome\Application\chrome.exe"
        self.config = BrowserConfig(
            execute_bin=canary,
            headless=False)

    def test_factory(self):
        factory = ChromeTabDevToolFactory(browser_config=self.config)
        chrome = BaseChrome(factory=factory)
        for i in range(10):
            r = chrome.get("https://baidu.com")
            # r = chrome.get("https://baidu.com")
            # r.window_size(1000, 1000)
            # r.scroll_by(0, 10000)
            print(r.execute_js("document.title"))
            print(r.text)
            print(r.cookies(urls=["baidu.com"]))
            # print(r.get_all_cookies())
            print(i)

    def test_pool(self):
        factory = ViewBrowserFactory(browser_config=self.config)
        pool = ResourcePool(factory=factory, init_count=2, max_count=4)
        threadpool = ThreadPoolExecutor(max_workers=5)

        def run_url(item):
            print("OK {}".format(item))
            start = time()
            c = pool.get()
            r = c.get("http://qq.com")
            pool.return_resource(c)
            print("end {}, {}".format(item, time() - start))

        for item in range(20):
            threadpool.submit(run_url, (item))

        while True:
            sleep(1)

    def test_multi_chrometab(self):
        cc = ChromeTab(browser_config=self.config)
        cc2 = ChromeTab(browser_config=self.config)
        r = cc.get("http://baidu.com")
        r2 = cc2.get("http://qq.com")
        print(r.text)


if __name__ == '__main__':
    main = TestBrowser()
    main.setUp()
    main.test_pool()



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/11/29
#
import atexit
import os
import platform
import subprocess
import threading
from time import sleep

import pychrome
import requests

from hunters.config import DEFAULT_BROWSER_CONFIG
from hunters.factory import Factory


def default_chrome_bin():
    if platform.system().lower() == "windows":
        return "chrome.exe"
    return "chrome"


def find_chrome_bin(bin_path):
    """
    find chrome bin location from
    the priority is
    env[CHROME_BIN] > bin_path > default_chrome_bin()
    :param bin_path:
    :return:
    """
    return os.getenv("CHROME_BIN", bin_path or default_chrome_bin())


def parse_chrome_option(browser_config, debugging_port=9222, debugging_address="127.0.0.1"):
    """
    parse browser_config to chrome config
    :param browser_config:
    :param debugging_address:9222 default
    :param debugging_port: 127.0.0.1 default
    :return: list
    """
    chrome_flag = [
        # '--headless',
        # '--disable-gpu',
        '--hide-scrollbars',
        '--disable-infobars',
        # '--start-fullscreen',
        '--start-maximized',
        # '--no-sandbox',
        '--window-size={},{}'.format(browser_config.width, browser_config.height),
        '--disable-client-side-phishing-detection',
        '--disable-background-networking',
        '--disable-web-security',
        '--allow-running-insecure-content',
        '--enable-experimental-web-platform-features',

        # r'--load-extension=""',
        # r'--disable-extensions-except=""'
    ]

    if debugging_address:
        chrome_flag.append("--remote-debugging-address={}".format(debugging_address))

    if debugging_port:
        chrome_flag.insert(0, "--remote-debugging-port={}".format(debugging_port))
        chrome_flag.append("--user-data-dir=/data/appdatas/chrome/{}".format(debugging_port))

    if browser_config.headless:
        chrome_flag.insert(0, "--headless")

    if browser_config.user_agent:
        chrome_flag.append("--user-agent=\"{}\"".format(browser_config.user_agent))

    if browser_config.disk_cache_size is not None and browser_config.disk_cache_size > 0:
        chrome_flag.append("--disk-cache-size={}".format(browser_config.disk_cache_size))

    if not browser_config.gpu:
        chrome_flag.insert(2, "--disable-gpu")

    chrome_flag += browser_config.extra_options  #: append extra options by user custom

    return chrome_flag


def open_chrome(port=9222, address="127.0.0.1", browser_config=DEFAULT_BROWSER_CONFIG):
    chrome_flags = parse_chrome_option(debugging_port=port, debugging_address=address, browser_config=browser_config)
    chrome_bin = find_chrome_bin(browser_config.execute_bin)
    chrome_flags.insert(0, chrome_bin)
    item = None
    try:
        item = subprocess.Popen(args=chrome_flags)  #: use list not string to FIX `File name too long`
    except FileNotFoundError as e:
        raise FileNotFoundError("can't found chrome bin, set env CHROME_BIN or use BrowserConfig to config before run")

    def close_subprocess():
        """ 主进程退出的时候自动退出这个Chrome进程 """
        item.kill()

    atexit.register(close_subprocess)
    MAX = 5
    count = 0
    while True:
        try:
            result = requests.get("http://{}:{}/json".format(address, port), timeout=5)
            print("[Hunters] check chrome open {}:{}".format(address, port))
            if result.ok:
                break
        except Exception as e:
            # wait for chrome openning
            sleep(1)
            pass

        count += 1
        if count > MAX:
            raise RuntimeError("Could not connect to chrome http://{}:{} ".format(address, port))

    return item


class ChromeDevToolFactory(Factory):
    """ open multi chrome instance, and bind multi debugging-port"""

    def __init__(self, browser_config, port=9222, address="127.0.0.1"):
        super().__init__(browser_config=browser_config)
        self._config = browser_config
        self._remote_browser_pool = []
        self._browser_process = []
        self._mutex = threading.Lock()
        self._tab_browser = {}
        self._address = address
        self._port = port

    def new_instance(self):
        with self._mutex:
            b_process = open_chrome(port=self._port, address=self._address,
                                    browser_config=self.config())
            self._browser_process.append(b_process)
            browser_ = pychrome.Browser(url="http://{}:{}".format(self._address, self._port))
            self._remote_browser_pool.append(browser_)
            self._port += 1

            tab_ = browser_.list_tab()[0]
            if hasattr(tab_, "start"):
                tab_.start()
            self._tab_browser[tab_.id] = {"browser": browser_, "process": b_process, "tab": tab_}
            return tab_  # return tab_ as a resource

    def destroy(self, obj):
        data_map = self._tab_browser.pop(obj.id)
        obj.stop()
        self.close_process(data_map)

    @classmethod
    def close_process(cls, data_map):
        if data_map:
            data_map.get("process").kill()
            del data_map

    def destroy_all(self, pool_list=None):
        for tab_, item in self._tab_browser.items():
            self.close_process(item)


class ChromeTabDevToolFactory(Factory):
    """ open only a Chrome, use multi tab as resource, when use new_instance
    this factory just open a chrome, and return a new Chrome Tab
    """

    def __init__(self, browser_config, port=9222, address="127.0.0.1"):
        super().__init__(browser_config=browser_config)
        self._address = address
        self._port = port
        self._browser_process = None
        self._init_chrome(browser_config)
        self._browser = pychrome.Browser(url="http://{}:{}".format(self._address, self._port))

    def _init_chrome(self, browser_config):
        if len(browser_config.manual_url) == 0:
            self._browser_process = open_chrome(port=self._port, address=self._address,
                                                browser_config=browser_config)
        else:
            addrs = browser_config.manual_url.split(":")
            self._address = addrs[0]
            self._port = addrs[1]

    def new_instance(self):
        """return a new chrome tab as resource """
        tab = self._browser.new_tab("about:blank")
        if hasattr(tab, "start"):  # [FIX] pychrome 0.2.0 start manual
            tab.start()
        return tab

    def destroy_all(self, pool_list=None):
        try:
            self._browser_process.kill()
        except Exception as e:
            # silent
            print(e)

    def destroy(self, obj):
        obj.stop()
        self._browser.close_tab(obj.id)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/9
from setuptools import setup, find_packages

setup(
    name='hunters',
    version='2.1.0',
    description="A spider automation framework, provide uniform API to use chrome Dev Protocol and requests",
    author='Qin Wei (ChineseTiger)',
    long_description='Hunters provide a uniform API to call Chrome DevTools  and requests for web crawlers',
    author_email='imqinwei@qq.com',
    url="https://github.com/LoginQin/hunters",
    license='Apache License, Version 2.0',
    packages=find_packages(),
    keywords='hunters spider headless crawlers screenshot',
    install_requires=['requests', 'lxml', 'cssselect', 'redis', 'pychrome==0.2.2', 'cachetools']
    # requests==2.18.1, lxml==3.8.0, cssselect==1.0.1, redis==2.10.5, cachetools==2.0.1
)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/12/19
#
import atexit
import logging
import os
import socket
from time import time


class SocketUtils:
    @classmethod
    def find_random_port(cls):
        """return a random free port"""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('', 0))
            return s.getsockname()[1]

    @classmethod
    def find_free_port(cls, base_port, socket_path):
        """start base_port to lookup a free port"""
        while base_port < 60000:
            if cls.is_open(ip="127.0.0.1", port=base_port):
                base_port += 1
            else:
                try:
                    cls.__check_socket("{}/{}".format(socket_path, base_port))
                    return base_port
                except FileExistsError as e:
                    base_port += 1

    @classmethod
    def __check_socket(cls, socket_file):
        """ use a file to check deny multi process open same port """
        if os.path.exists(socket_file) and time() - os.stat(socket_file).st_mtime < 10:
            # file exits and last access less then 10s
            raise FileExistsError("socket exists")

        with open(socket_file, mode="w") as file:
            file.write(str(os.getpid()))
        atexit.register(cls.__remove_file, socket_file)

    @classmethod
    def __remove_file(cls, socket_file):
        try:
            print("DELETE {}".format(socket_file))
            os.remove(socket_file)
        except Exception as e:
            logging.error(e)

    @staticmethod
    def is_open(ip, port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((ip, int(port)))
            s.shutdown(2)
            # 利用shutdown()函数使socket双向数据传输变为单向数据传输。shutdown()需要一个单独的参数，
            # 该参数表示了如何关闭socket。具体为：0表示禁止将来读；1表示禁止将来写；2表示禁止将来读和写。
            s.close()
            return True
        except:
            return False
